/*------------------------------------------------------------------------*/
/* Copyright (C) 2015 Armin Biere Johannes Kepler University Linz Austria */
/*------------------------------------------------------------------------*/

// This is actually a template file for 'forward', 'backward', 'middle'.

/*------------------------------------------------------------------------*/
#if defined (STATICVISITED)

// The visited bit-map can be allocated statically for LC=9,LV=7 (512MB)
// but generates a "'.bss' too large error" for larger values.  It seems
// that static allocation in the BSS part gives 10% faster running times.

#if LC > 9
#warning "LC probably too large for '-s'"
#endif
#if LV > 7
#warning "LV probably too large for '-s'"
#endif

static unsigned visited[RC*RC*RV*RR];

#else

static unsigned * visited;

#endif
/*------------------------------------------------------------------------*/

// (signed) velocity to unsigned int translation
#define V2U(V) \
  (assert (MINV <= (V)), \
   assert ((V) <= MAXV), \
   ((unsigned)((V) - MINV)))

static void mark_visited (unsigned x, unsigned y, int w, int h) {
  visited[
    (     x << (LC + LV + LR)) |
    (     y << (     LV + LR)) |
    (V2U(w) << (          LR)) |
    (V2U(h) >> 5)]
  |=
    1u<<(V2U(h)&31)
  ;
}

static unsigned has_been_visited (unsigned x, unsigned y, int w, int h) {
  unsigned res;
  assert (can_handle_state (x, y, w, h));
  lookups++;
  res = visited[
    (     x << (LC + LV + LR)) |
    (     y << (     LV + LR)) |
    (V2U(w) << (          LR)) |
    (V2U(h) >> 5)]
  & 
    (1u<<(V2U(h)&31))
  ;
  return res;
}

/*------------------------------------------------------------------------*/
#ifndef ONLYVISITED
/*------------------------------------------------------------------------*/

State * states;					// stack of states
size_t nstates;					// states on stack

/*------------------------------------------------------------------------*/

static size_t szstates;				// size of stack
static Gen * gens;				// breadth-first search
static size_t ngens, szgens;			// generation stack

/*------------------------------------------------------------------------*/

static void enlarge_states () {
  size_t old_bytes = szstates * sizeof *states, new_bytes;
  if (!szstates) szstates = 1;
  else if (szstates * sizeof *states >= (1l<<30))	// at least 1 GB
    szstates += (1l<<30) / sizeof *states;		// then inc by 1 GB
  else szstates *= 2;					// otherwise double
  new_bytes = szstates * sizeof *states;	
  LOCK_STATES ();
  states = reallocate (states, old_bytes, new_bytes);
  UNLOCK_STATES ();
  if (new_bytes >= (1l << 30))
    msg ("%.02f: enlarged %s stack to hold %lu states %.0f GB",
      get_time () - started, TYPE_GENERATION,
      szstates, new_bytes /(double)(1<<30));
}

static int statex (unsigned state) {
  if (state < nstates) return states[state].x;
  assert (!state), assert (!nstates);
  return start[0];
}

static int statey (unsigned state) {
  if (state < nstates) return states[state].y;
  assert (!state), assert (!nstates);
  return start[1];
}

static int deltax (unsigned state) {
  assert (state < nstates);
  return statex (state) - statex (states[state].pred);
}

static int deltay (unsigned state) {
  assert (state < nstates);
  return statey (state) - statey (states[state].pred);
}

static unsigned push_state (int x, int y, int w, int h, unsigned pred) {
  unsigned res = nstates;
  State state = { x, y, pred };
  assert (state.x == x), assert (state.y == y);
  if (nstates == szstates) enlarge_states ();
  states[nstates] = state;
  LOCK_STATES ();
  nstates++;
  if (nstates > max_active_states) max_active_states = nstates;
  UNLOCK_STATES ();
  LOGMSG (TYPE_GENERATION " STATE[%u] %d %d %d %d PRED %u",
    res, x, y, w, h, pred);
  assert (deltax (res) == w), assert (deltay (res) == h);
  pushed_states++;
  if (!(unsigned)nstates) {
    if (garbage_collection_enabled)
      die ("too many active states (even with garbage collection)");
    else
      die ("too many active states (might want to use '-g')");
  }
  return res;
}

/*------------------------------------------------------------------------*/

static void enlarge_generations () {
  size_t old_bytes = szgens * sizeof *gens, new_bytes;
  szgens = szgens ? 2*szgens : 1;
  new_bytes = szgens * sizeof *gens;
  gens = reallocate (gens, old_bytes, new_bytes);
}

/*------------------------------------------------------------------------*/

static void garbage_collection (Gen * g) {
  size_t szmarkbits, map_bytes, map_size, pos, orig_nstates, new_front;
  size_t marked = 0, unmarked = 0, moved = 0, collected = 0;
  double mark_time, collect_time, total_time;
  double begin = get_time (), middle, end;
  size_t old_begin, new_begin;
  unsigned * markbits, * map;
  Gen * h;
  size_t s, p;
  assert (garbage_collection_enabled);

  // Need at least three, since first and last ('g') are kept.
  if (ngens < 3) return;

  /*----------------------------------------------------------------------*/

  // We will mark only states before last generation 'g'.
  assert (g->begin);
  szmarkbits = g->begin;
  while (szmarkbits & 31) szmarkbits++;
  szmarkbits >>= 5;
  markbits = callocate (szmarkbits * sizeof *markbits);

  // So first mark predecessors of last generation.
  for (s = g->end - 1; s >= g->begin; s--) {
    p = states[s].pred;
    MARKBITIFNOTMARKED (p);
  }

  // Then go backward and mark all recursively (in BFS).
  assert (s == g->begin - 1);
  do {
    if (MARKEDBIT (s)) {
      p = states[s].pred;
      MARKBITIFNOTMARKED (p);
    } else unmarked++;
  } while (s--);

  middle = get_time ();
  mark_time = middle - begin;
#if 0
  msg ("%.2f: "
    "marked %s %lu states %.0f%% in %.03f sec",
    middle - started, TYPE_GENERATION,
    marked, percent (marked, nstates), mark_time);
  assert (marked + unmarked + g->last - g->first + 1 == nstates);
#endif

  /*----------------------------------------------------------------------*/

  // Now we will compactify and move states starting from second
  // generation.  In order to fix the predecessor links in the
  // next generation we keep a map of old indices to new indices.
  // Since we can move and fix generation by generation, we only
  // need a map as big as the largest generation before 'g'.
  map_size = 0;
  for (h = gens; h < g; h++) {
    size_t tmp = size_generation (h);
    if (tmp > map_size) map_size = tmp;
  }
  // msg ("map size %lu", map_size);
  map_bytes = map_size * sizeof *map;
  map = allocate (map_bytes);

  assert (ngens > 1);
  moved = gens[0].end;
  assert (moved == gens[1].begin);

  new_front = 0;

  LOCK_STATES (); /*------------------------ start mutating states -------*/

  // For all generations except first and last.
  for (h = gens + 1; h < g; h++) {

    // First, move and map old position to new in 'h[0]'.
    new_begin = moved;
    old_begin = h->begin;

    for (s = h->begin; s < h->end; s++) {
      unsigned smarked = MARKEDBIT (s);
      if (smarked) {
	assert (moved <= s);
	states[moved] = states[s];
	pos = s - old_begin;
	assert (pos < map_size);
	map[pos] = moved++;
      } else collected++;
#ifdef VISUALIZE
      if (visualize) {
	if (forward_last_visualized == s) forward_last_visualized = moved-1;
#ifdef BACKWARD
	if (backward_last_visualized == s) backward_last_visualized = moved-1;
#endif
      }
#endif
    }

    // Second, fix predcessor links in next 'h[1]'.
    for (s = h[1].begin; s < h[1].end; s++) {
      if (h + 1 < g && !MARKEDBIT (s)) continue;
      p = states[s].pred;
      assert (old_begin <= p);
      assert (p < h->end);
      pos = p - old_begin;
      assert (pos < map_size);
      states[s].pred = map[pos];
    }

    h->begin = new_begin;
    h->end = moved;
  }

  // Finally move last generation 'g'.
  new_begin = moved;
  for (s = g->begin; s < g->end; s++) {
    assert (moved <= s);
    states[moved] = states[s];
#ifdef VISUALIZE
    if (visualize) {
      if (forward_last_visualized == s) forward_last_visualized = moved-1;
#ifdef BACKWARD
      if (backward_last_visualized == s) backward_last_visualized = moved-1;
#endif
    }
#endif
    moved++;
  }
  orig_nstates = nstates;
  if (!new_front) {
    new_front = moved - 1;
  }
  g->begin = new_begin;
  g->end = moved;
  nstates = moved;

  UNLOCK_STATES (); /*----------------------- stop mutating states -------*/

  /*----------------------------------------------------------------------*/

#ifndef NDEBUG
  if (MARKEDBIT (0)) {
    assert (collected == unmarked);
    assert (moved == marked + g->end - g->begin);
  }
#endif
  deallocate (markbits, szmarkbits * sizeof *markbits);
  deallocate (map, map_bytes);
  end = get_time ();
  collect_time = end - middle;
  total_time = mark_time + collect_time;
  garbage_collection_time += total_time;
#if 0
  msg ("%.2f: "
    "collected %s %lu states %.0f%% in %.03f sec %.0f%%",
    end - started, TYPE_GENERATION,
    collected, percent (marked, orig_nstates), collect_time,
    percent (collect_time, total_time));
#endif
  msg ("%.2f: "
    "marked %s %lu collected %s %.0f%% in %.03f sec",
    end - started, TYPE_GENERATION, g - gens,  
    b2s (collected*sizeof*states),
    percent (collected, orig_nstates),
    total_time);
}

/*------------------------------------------------------------------------*/

static Gen * last_generation () {
  assert (ngens);
  return gens + (ngens - 1);
}

static size_t end_of_last_generation () {
  return ngens ? last_generation ()->end : 0;
}

static void push_generation () {
  Gen * g;
  if (ngens == szgens) enlarge_generations ();
  g = gens + ngens;
  g->begin = end_of_last_generation ();
  g->end = nstates;
  LOGMSG (TYPE_GENERATION " GENERATION %lu [%lu,%lu] %lu states",
    ngens, g->begin, g->end - 1, size_generation (g));
  if (!do_not_sort_generations)
    sort_generation (TYPE_GENERATION, states, gens, g);
  if (garbage_collection_enabled) garbage_collection (g);
  ngens++;
}

/*------------------------------------------------------------------------*/

static unsigned new_state (int x, int y, int w, int h, unsigned pred) {
  unsigned res;
  assert (is_valid_point (x, y));
  assert (can_handle_state (x, y, w, h));
  assert (!((x-START[0]) % incaccel));
  assert (!((y-START[1]) % incaccel));
  assert (!(w % incaccel));
  assert (!(h % incaccel));
  if (w < min_delta_x) min_delta_x = w;
  if (h < min_delta_y) min_delta_y = h;
  if (w > max_delta_x) max_delta_x = w;
  if (h > max_delta_y) max_delta_y = h;
  res = push_state (x, y, w, h, pred);
  mark_visited (x, y, w, h);
#ifndef NLOG
  if (is_logging_enabled ()) {
    if (res) {
      const int fx = statex (pred), fy = statey (pred);
      const int fw = deltax (pred), fh = deltay (pred);
      const int tx = statex (res), ty = statey (res);
      const int tw = deltax (res), th = deltay (res);
      const int aw = tw - fw, ah = th - fh;
      assert (w == tw), assert (h == th);
      msg (TYPE_GENERATION 
        " TRANSITION %d %d %d %d --[%d,%d]-> %d %d %d %d (%d,%d)",
	fx, fy, fw, fh,
	tw, th,
	tx, ty, tw, th,
	aw, ah);
    } else msg (TYPE_GENERATION " START %d %d %d %d", x, y, w, h);
  }
#endif
  return res;
}

static int expand_last_generation (unsigned * state_idx_ptr) {
#ifdef VISUALIZE
  const int check_count_down_interval = 1000;
  int check_count_down = check_count_down_interval;
#endif
  double entered = get_time (), t, d;
  Gen * g = last_generation ();
  unsigned idx;
  for (idx = g->begin; idx < g->end; idx++) {
    const int fx = statex (idx), fy = statey (idx);
    const int fw = deltax (idx), fh = deltay (idx);
    const int * a;
#ifdef VISUALIZE
    if (visualize && !--check_count_down) {
      if (CHECK_DONE ()) return -1;
      check_count_down = check_count_down_interval;
    }
#endif
    LOGMSG (TYPE_GENERATION " FRONTIER %d %d %d %d", fx, fy, fw, fh);
    for (a = accels; a < eoa; a += 2) {
      const int w = fw + a[0], h = fh + a[1];
      const int x = fx + w, y = fy + h;
      LOGMSG (TYPE_GENERATION " ACCEL %d %d DELTA %d %d",
        a[0], a[1], w, h);
      generated++;
      if (can_handle_state (x, y, w, h)) {
	if (has_been_visited (x, y, w, h)) hits++;
	else if (is_valid_move (fx, fy, x, y)) {
	  const unsigned state_idx = new_state (x, y, w, h, idx);
	  if (target_reached (x, y, w, h)) {
	    *state_idx_ptr = state_idx;
	    return 0;
	  }
	  new_reachable_state (x, y, w, h);
	} else mark_visited (x, y, w, h);
      } else if (is_valid_move (fx, fy, x, y))
	report_unhandled_state (x, y, w, h);
    }
  }
  t = get_time ();
  d = t - entered;
  msg ("%.2f: expand %s %lu of %lu states in %.3f sec",
    t - started, TYPE_GENERATION, g - gens, size_generation (g), d);
  expand_time += d;
  push_generation ();
  return 1;
}

/*------------------------------------------------------------------------*/
#endif // ONLYVISITED
/*------------------------------------------------------------------------*/

