/*------------------------------------------------------------------------*/
/* Copyright (C) 2015 Armin Biere Johannes Kepler University Linz Austria */
/*------------------------------------------------------------------------*/

#ifdef VISUALIZE

/*------------------------------------------------------------------------*/

#include "common.h"
#include "visualize.h"
#include "state.h"

/*------------------------------------------------------------------------*/

extern State * forward_states;
extern size_t forward_nstates;

#ifdef BACKWARD
extern State * backward_states;
extern size_t backward_nstates;
#endif

/*------------------------------------------------------------------------*/

#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glu.h>

extern void glutLeaveMainLoop ();
extern void glutBitmapString (void *, const char *);

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>

/*------------------------------------------------------------------------*/

static float scale = 2, W, H, B, A, S, T;

static int show_grid;
static int full_screen;
static int keep_aspect_ratio = 1;
static int show_velocity = 1;
static int thickness = 1;
static int paths = 100;
static float show_keys;
static int inverted;

static int dec_show_keys_triggered;

static int playing, paused;
static float play_delay = 1.0 / 20.0;

/*------------------------------------------------------------------------*/

unsigned forward_last_visualized;
unsigned backward_last_visualized;

/*------------------------------------------------------------------------*/

static pthread_t thread;
static int states_locked;
static pthread_mutex_t states_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t done_mutex = PTHREAD_MUTEX_INITIALIZER;
static int initialized, done;

/*------------------------------------------------------------------------*/

void lock_states () { 
  pthread_mutex_lock (&states_mutex);
  assert (!states_locked);
  states_locked = 1;
}

void unlock_states () { 
  assert (states_locked);
  states_locked = 0;
  pthread_mutex_unlock (&states_mutex);
}

static unsigned forward_end_state () {
  assert (states_locked);
  return forward_nstates ? forward_nstates - 1 : 0;
}

#ifdef BACKWARD
static unsigned backward_end_state () {
  assert (states_locked);
  return backward_nstates ? backward_nstates - 1 : 0;
}
#endif

static void mark_done () {
  pthread_mutex_lock (&done_mutex);
  done = 1;
  pthread_mutex_unlock (&done_mutex);
}

int check_done () {
  int res;
  pthread_mutex_lock (&done_mutex);
  res = done;
  pthread_mutex_unlock (&done_mutex);
  return res;
}

/*------------------------------------------------------------------------*/

static const char * keys[][2] = {
  { "q",                "quit application" },
  { "f",                "toggle full screen" },
  { "g",                "toggle grid" },
  { "t",                "toggle point & line thickness"},
  { "i",                "invert foreground and background colors"},
  { "p",                "start continous playing"},
  { "j/k",              "increase/decrease playing speed"},
  { "v",                "toggle showing velocity" },
  { "a",                "toggle keeping aspect ratio" },
  { "space key",        "toggle pausing" },
  { "left/right arrow", "slow forward / backward (1 step)" },
  { "up/down arrow",    "medium forward / backward (100 steps)" },
  { "page-up/down",     "fast forward / backward (10000 steps)" },
  { "home, end",        "go to start / end position" },
  { "0,1,2,3,4,...,7",   "show 1,10,100,...,1e6,1e7 paths" },
  { "<other-key>",      "show this key usage help screen" }
}
;

static const int nkeys = sizeof keys / sizeof *keys;

/*------------------------------------------------------------------------*/

static float bg () { return inverted ? 1.0 : 0.0; }
static float fg () { return inverted ? 0.0 : 1.0; }

static void vertex (float a, float b) { glVertex2f (a, b); }

static void delay_inc_last ();

static void inc_last (int inc) {
  int continue_delaying = 0;
  unsigned long end;

  if (!playing) return;

  lock_states ();

  if (forward_states) {
    end = forward_end_state ();
    if (!paused) {
      if (UINT_MAX - inc < forward_last_visualized)
	forward_last_visualized = end;
      else if (forward_last_visualized + inc <= end)
	forward_last_visualized += inc;
      else
	forward_last_visualized = end;
    }
    if (forward_last_visualized < end) continue_delaying = 1;
  }

#ifdef BACKWARD
  if (backward_states) {
    end = backward_end_state ();
    if (!paused) {
      if (UINT_MAX - inc < backward_last_visualized)
	backward_last_visualized = end;
      else if (backward_last_visualized + inc <= end)
	backward_last_visualized += inc;
      else
	backward_last_visualized = end;
    }
    if (backward_last_visualized < end) continue_delaying = 1;
  }
#endif

  unlock_states ();

  if (continue_delaying) delay_inc_last ();
  else playing = 0;
}

static void delay_inc_last () {
  float delay, inc;
  if (play_delay < 0.01) delay = 0.01, inc = 0.01 / play_delay;
  else delay = play_delay, inc = 1;
  glutTimerFunc (1000*delay, inc_last, inc);
}

static unsigned display_transitions (State * states, unsigned end) {
  unsigned begin, i, j, pred, transitions;
  float delta;

  begin = end >= paths ? end - paths + 1: 1;
  delta = end - begin + 1;

  for (i = begin; i < end; i++) {
    float f = (i - begin) / delta;
    float g = f > 0.5 ? f - 0.5 : 0;
    glBegin (GL_LINES);
    glColor3f (g, g, f);
    for (j = i; j; j = pred) {
      pred = states[j].pred;
      vertex (states[pred].x, states[pred].y);
      vertex (states[j].x, states[j].y);
    }
    glEnd ();
    glPointSize (2 * thickness);
    glBegin (GL_POINTS);
    glColor3f (0, f, 1);
    for (j = i; j; j = states[j].pred)
      vertex (states[j].x, states[j].y);
    glEnd ();
  }

  glPointSize (6 * thickness);
  glColor3f (0, 1, 0);
  glBegin (GL_POINTS);
  vertex (start[0], start[1]);
  glEnd ();

  glColor3f (1, 0, 0);
  glBegin (GL_POINTS);
  vertex (goal[0], goal[1]);
  glEnd ();

  glColor3f (0, .75, 0);
  glRasterPos2i(start[0], start[1]);
  glutBitmapString (GLUT_BITMAP_HELVETICA_12, "  start");

  glColor3f (.75, 0, 0);
  glRasterPos2i(goal[0], goal[1]);
  glutBitmapString (GLUT_BITMAP_HELVETICA_12, "  goal");

  transitions = 0;
  glLineWidth (2.0 * thickness);
  glBegin (GL_LINES);
  for (j = end; j; j = pred) {
    transitions++;
    pred = states[j].pred;
    if (pred && show_velocity) {
      unsigned predpred = states[pred].pred;
      int pw = states[pred].x - states[predpred].x;
      int ph = states[pred].y - states[predpred].y;
      int vx = states[pred].x + pw;
      int vy = states[pred].y + ph;
      glEnd ();
      glLineWidth (1.5 * thickness);
      glBegin (GL_LINES);
      glColor3f (0, .5, 0);
      vertex (states[pred].x, states[pred].y);
      vertex (vx, vy);
      glColor3f (.5, 0, 0);
      vertex (vx, vy);
      vertex (states[j].x, states[j].y);
      glEnd ();
      glLineWidth (2 * thickness);
      glBegin (GL_LINES);
    }
    if (inverted) glColor3f (.75, .75, 0);
    else glColor3f (1, 1, 0);
    vertex (states[pred].x, states[pred].y);
    vertex (states[j].x, states[j].y);
  }
  glEnd ();
  glLineWidth (1 * thickness);

  glPointSize (4 * thickness);
  glBegin (GL_POINTS);
  glColor3f (fg (), fg (), fg ());
  for (j = end; j; j = states[j].pred)
    vertex (states[j].x, states[j].y);
  glEnd ();

  glPointSize (6 * thickness);
  glBegin (GL_POINTS);
  glColor3f (fg (), fg (), fg ());
  vertex (states[end].x, states[end].y);
  glEnd ();

  return transitions;
}

static void display() {
  unsigned end, transitions;
  char buffer[400];
  const int * p;

  glClearColor (bg (), bg (), bg (), 0);
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  if (show_grid) {
    int i;
    glLineWidth (.25 * thickness);
    if (inverted) glColor3f (.9, .9, .9);
    else glColor3f (.1,.1,.1);
    glBegin (GL_LINES);
    for (i = box[0]; i <= box[2]; i++)
      vertex (i, box[1]), vertex (i, box[3]); 
    for (i = box[1]; i <= box[3]; i++)
      vertex (box[0], i), vertex (box[2], i); 
    glEnd ();
  }

  glLineWidth (1 * thickness);

  glColor3f (0, 1, 1);
  glBegin (GL_LINES);
  vertex (box[0], box[1]), vertex (box[0], box[3]), 
  vertex (box[0], box[3]), vertex (box[2], box[3]),
  vertex (box[2], box[3]), vertex (box[2], box[1]),
  vertex (box[2], box[1]), vertex (box[0], box[1]);
  glColor3f (fg (), fg (), fg ());
  for (p = segs; p < top_segs; p += 2) vertex (p[0], p[1]);
  glEnd ();

  lock_states ();

  buffer[0] = 0;

  if (forward_states) {
    if (paused || playing) end = forward_last_visualized;
    else end = forward_end_state ();
    transitions = display_transitions (forward_states, end);
#ifdef BACKWARD
    if (backward_states) strcat (buffer, "forward ");
#endif
    sprintf (buffer + strlen (buffer),
      "%u states %u transitions", end + 1, transitions);
#ifdef BACKWARD
    if (backward_states) strcat (buffer, ", ");
#endif
  }

#ifdef BACKWARD
  if (backward_states) {
    if (paused || playing) end = backward_last_visualized;
    else end = backward_end_state ();
    transitions = display_transitions (backward_states, end);
    if (forward_states) strcat (buffer, "backward ");
    sprintf (buffer + strlen (buffer),
      "%u states %u transitions", end + 1, transitions);
  }
#endif

  unlock_states ();


  if (paused) strcat (buffer, ", paused");
  if (playing)
      sprintf (buffer + strlen (buffer),
         ", playing %.1f/sec", 1/play_delay);
  if (done) strcat (buffer, ", done.");
  else strcat (buffer, ", ...");

  glColor3f (fg (), fg (), fg ());
  glRasterPos2i(1, 1);
  glutBitmapString (GLUT_BITMAP_HELVETICA_12, buffer);

  if (show_keys > 0.1) {
    float x = 2, y = H, w, h, f;
    int i;
    w = 6.0 * (float) S / (float) glutGet (GLUT_WINDOW_WIDTH);
    h = 16.0 * (float) T / (float) glutGet (GLUT_WINDOW_HEIGHT);
    f = show_keys > 1 ? 1 : show_keys;
    if (inverted) f = 1.0 - f;
    glColor3f (f, f, f);
    for (i = 0; i < nkeys; i++) {
      y -= h;
      glRasterPos2i(x, y);
      glutBitmapString (GLUT_BITMAP_HELVETICA_12, keys[i][0]);
      glRasterPos2i(x + 18.0*w, y);
      glutBitmapString (GLUT_BITMAP_HELVETICA_12, keys[i][1]);
    }
  } else if (show_keys) show_keys = 0;

  glFlush ();
  glutSwapBuffers ();
  glutPostRedisplay ();
}

static int savedw, savedh;

static void reshape (int w, int h) {
  float f = 0, g = 0;
  glViewport (0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  if (keep_aspect_ratio && w && h && A) {
    float a = w / (float) h;
    if (a) {
	   if (a < A) g = (h - w/A)*((float)W/w)/2.0;
      else if (a > A) f = (w - h*A)*((float)H/h)/2.0;
    }
  }
  S = W + 2*f, T = H + 2*g;
  gluOrtho2D (-f, W + f, -g, H + g);
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.375, 0.375, 0.0);
}

static void set_last_visualized () {
  if (forward_states)
    forward_last_visualized = forward_end_state ();
#ifdef BACKWARD
  if (backward_states)
    backward_last_visualized = backward_end_state ();
#endif
}

static void pause_visualization () {
  if (!playing) {
    lock_states ();
    set_last_visualized ();
    unlock_states ();
  }
  paused = 1;
}

static void dec_show_keys (int dummy) {
  (void) dummy;
  assert (dec_show_keys_triggered);
  if (show_keys > 1) {
    show_keys -= 1;
    glutTimerFunc (1000, dec_show_keys, 0);
  } else if (show_keys > 0) {
    show_keys -= 0.0025;
    if (show_keys > 0) glutTimerFunc (10, dec_show_keys, 0);
    else show_keys = 0;
  } 
  if (show_keys < 0) show_keys = 0;
  if (!show_keys) dec_show_keys_triggered = 0;
}

static void toggle_show_keys () { 
  if (show_keys > 1) show_keys = 0;
  else {
    show_keys = 3;
    if (!dec_show_keys_triggered) {
      dec_show_keys_triggered = 1;
      glutTimerFunc (1000, dec_show_keys, 0);
    }
  }
}

static void continues_playing () {
  if (paused) paused = 0;
  else {
    lock_states ();
    set_last_visualized ();
    unlock_states ();
  }
  playing = 1;
  delay_inc_last ();
}

static void key (unsigned char k, int x, int y) { 
  if (k == 'q') {
    if (check_done ()) glutLeaveMainLoop ();
    else mark_done ();
  } else  if (k == 'a') {
    keep_aspect_ratio = !keep_aspect_ratio;
    reshape (glutGet (GLUT_WINDOW_WIDTH), glutGet (GLUT_WINDOW_HEIGHT));
  } else  if (k == 'f') {
    if (full_screen) {
      full_screen = 0;
      glutReshapeWindow (savedw, savedh);
    } else {
      savedw = glutGet (GLUT_WINDOW_WIDTH);
      savedh = glutGet (GLUT_WINDOW_HEIGHT);
      full_screen = 1;
      glutFullScreen ();
    }
  } else  if (k == 'g') {
    show_grid = !show_grid;
  } else if (k == 'v') {
    show_velocity = !show_velocity;
  } else if (k == 'p') {
    if (playing) playing = 0;
    else continues_playing ();
  } else if (k == 't') {
    thickness = 1 + (thickness == 1);
  } else if (k == 'i') {
    inverted = !inverted;
  } else if (k == 'k') {
    play_delay *= 1.1;
    if (play_delay > 1) play_delay = 1;
  } else if (k == 'j') {
    play_delay *= .9;
    if (play_delay < 0.000001) play_delay = 0.000001;
  } else if (k == ' ') {
    if (paused) paused = 0;
    else pause_visualization ();
  }
  else if (k == '0') paths = 1;
  else if (k == '1') paths = 10;
  else if (k == '2') paths = 100;
  else if (k == '3') paths = 1000;
  else if (k == '4') paths = 10000;
  else if (k == '5') paths = 100000;
  else if (k == '6') paths = 1000000;
  else if (k == '7') paths = 10000000;
  else toggle_show_keys ();
}

static void reverse (unsigned dec) {
  if (!paused) pause_visualization ();
  lock_states ();
  if (forward_states) {
    if (forward_last_visualized >= dec && forward_last_visualized - dec > 1)
      forward_last_visualized -= dec;
    else forward_last_visualized = 1;
  }
#ifdef BACKWARD
  if (backward_states) {
    if (backward_last_visualized >= dec && backward_last_visualized - dec > 1)
      backward_last_visualized -= dec;
    else backward_last_visualized = 1;
  }
#endif
  unlock_states ();
}

static void forward (unsigned inc) {
  unsigned long end;
  if (!paused) pause_visualization ();
  lock_states ();
  if (forward_states) {
    end = forward_end_state ();
    if (UINT_MAX - inc < forward_last_visualized)
      forward_last_visualized = end;
    else if (forward_last_visualized + inc <= end)
      forward_last_visualized += inc;
    else forward_last_visualized = end;
  }
#ifdef BACKWARD
  if (backward_states) {
    end = backward_end_state ();
    if (UINT_MAX - inc < backward_last_visualized)
      backward_last_visualized = end;
    else if (backward_last_visualized + inc <= end)
      backward_last_visualized += inc;
    else backward_last_visualized = end;
  }
#endif
  unlock_states ();
}

static void special_key (int k, int x, int y) { 
       if (k == GLUT_KEY_LEFT) reverse (1);
  else if (k == GLUT_KEY_RIGHT) forward (1);
  else if (k == GLUT_KEY_DOWN) forward (100);
  else if (k == GLUT_KEY_UP) reverse (100);
  else if (k == GLUT_KEY_PAGE_DOWN) forward (10000);
  else if (k == GLUT_KEY_PAGE_UP) reverse (10000);
  else if (k == GLUT_KEY_END) forward (UINT_MAX);
  else if (k == GLUT_KEY_HOME) reverse (UINT_MAX);
  else toggle_show_keys ();
}

static void * run_opengl_thread (void * dummy) {
  char * argv[] = { "mcracetrack" };
  int argc = 1;
  B = 4;
  W = box[2] + 1;
  H = box[3] + 1;
  S = W, T = H;
  assert (H > 0);
  assert (W > 0);
  A = W / H;
  glutInit (&argc, argv);
  glutInitDisplayMode (GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  glutInitWindowSize (scale*W, scale*H);
  glutCreateWindow ("racetrack");
  glutDisplayFunc (display);
  glutReshapeFunc (reshape);
  glutKeyboardFunc (key);
  glutSpecialFunc (special_key);
  glutSetCursor(GLUT_CURSOR_NONE); 
  glutMainLoop ();
  return dummy;
}

void init_visualization () {
  int i;
  msg ("");
  msg ("initializing visualization, use the following keys:");
  msg ("");
  for (i = 0; i < nkeys; i++) msg ("  %-20s %s", keys[i][0], keys[i][1]);
  msg ("");
  initialized = 1;
  pthread_create (&thread, 0, run_opengl_thread, 0);
}

void reset_visualization () {
  if (!initialized) return;
  if (check_done ()) glutLeaveMainLoop ();
  else mark_done ();
  pthread_join (thread, 0);
}

#endif
